# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Canary helper code
"""

import logging


log = logging.getLogger(__name__)


def get_repo_info():
    """[EXPERIMENTAL] Obtains info on the current repo the code is in.

    Returns:
        dict: git meta data"""
    try:
        import git

        repo = git.Repo(search_parent_directories=True)
        branch = repo.active_branch
        head = repo.head
        return {
            "git": repo.remotes.origin.url,
            "branch": branch.name,
            "commit": head.commit.hexsha,
            "last_known_author": head.commit.author.name.encode(
                "ascii", "ignore"
            ).decode(),
        }
    except:
        return {"git": "n/a"}


def test_pipeline_step_metrics(pipeline_run, expected_metrics):
    """Tests a pipeline run against a set of expected metrics.

    Args:
        pipeline_run (PipelineRun): the AzureML pipeline run
        expected_metrics (dict): defines the tests to execute

    Returns:
        List: errors collected during tests

    Notes:
        example entries in expected_metrics

        "SelectJsonField" : [{"row" : {"name" : "output", "key" : "size", "value" : 369559}}],
        tests module "SelectJsonField" for a metric row named "output", checks key "size" must have value 369559

        "tokenizerparallel" : [{"metric" : {"key" : "Failed Items", "value" : 0}}],
        tests module "tokenizerparallel" for a metric named "Failed Items", value must be 0
    """
    errors = []

    log.info("Looping through PipelineRun steps to test metrics...")
    for step in pipeline_run.get_steps():
        log.info(f"Checking status of step {step.name}...")

        observed_metrics = step.get_metrics()
        log.info(f"Step Metrics: {observed_metrics}")

        status = step.get_status()
        if status != "Finished":
            errors.append(f"Pipeline step {step.name} status is {status} != Finished")

        if step.name in expected_metrics:
            for expected_metric_test in expected_metrics[step.name]:
                if "row" in expected_metric_test:
                    log.info(f"Checking metrics, looking for {expected_metric_test}")
                    row_key = expected_metric_test["row"]["name"]
                    metric_key = expected_metric_test["row"]["key"]
                    expected_value = expected_metric_test["row"]["value"]
                    if row_key not in observed_metrics:
                        errors.append(
                            f"Step {step.name} metric row '{row_key}' not available in observed metrics {observed_metrics}"
                        )
                    elif metric_key not in observed_metrics[row_key]:
                        errors.append(
                            f"Step {step.name} metric row '{row_key}' does not have a metric '{metric_key}' in observed metrics {observed_metrics[row_key]}"
                        )
                    elif observed_metrics[row_key][metric_key] != expected_value:
                        errors.append(
                            f"Step {step.name} metric row '{row_key}' - metric '{metric_key}' - does not have expected value {expected_value} in observed metrics {observed_metrics[row_key]}"
                        )
                if "metric" in expected_metric_test:
                    log.info(f"Checking metrics, looking for {expected_metric_test}")
                    metric_key = expected_metric_test["metric"]["key"]
                    expected_value = expected_metric_test["metric"]["value"]
                    if metric_key not in observed_metrics:
                        errors.append(
                            f"Step {step.name} metric '{metric_key}' not available in observed metrics {observed_metrics}"
                        )
                    elif observed_metrics[metric_key] != expected_value:
                        errors.append(
                            f"Step {step.name} metric row '{metric_key}' does not have expected value {expected_value} in observed metrics {observed_metrics[metric_key]}"
                        )

    return errors


def test_pipeline_step_output(pipeline_run, step_name, output_name, **kwargs):
    """Verify a given pipeline output for some basic checks.

    Args:
        pipeline_run (PipelineRun): the pipeline run
        step_name (str): name of the step to check
        output_name (str): name of the output to check
        **kwargs: Arbitrary keyword arguments defining the test

    Kwargs:
        length (int) : to verify the length

    Returns:
        dict: results
    """
    pipeline_step = pipeline_run.find_step_run(step_name)
    results = {"errors": []}

    if not pipeline_step:
        results["exception"] = (
            f"Could not find step {step_name} in pipeline {pipeline_run._run_id}."
        )
        return results

    output_port = pipeline_step[0].get_output_data(output_name)
    if not output_port:
        results["exception"] = (
            f"Could not find output {output_name} in step {step_name} in pipeline {pipeline_run._run_id}."
        )
        return results

    data_reference = output_port._data_reference

    data_path = DataPath(
        datastore=data_reference.datastore,
        path_on_datastore=data_reference.path_on_datastore,
        name=data_reference.data_reference_name,
    )

    if kwargs.get("length"):
        expected_length = kwargs.get("length")
        log.info(
            f"Checking count={expected_length} of files for step {step_name} output {output_name}..."
        )
        data_set = Dataset.File.from_files(data_path)

        files_list = data_set.to_path()

        if expected_length < 0:
            # test any length > 0
            results["length"] = {"expected": ">0", "observed": len(files_list)}
            if results["length"]["observed"] == 0:
                message = """Length mismatch in output {output_name} in step {step_name} in pipeline {run_id}. Expected len {a} found {b}.""".format(
                    output_name=output_name,
                    step_name=step_name,
                    run_id=pipeline_run._run_id,
                    b=results["length"]["observed"],
                    a=results["length"]["expected"],
                )
                # logging.error(message)
                results["errors"].append(message)
        else:
            results["length"] = {
                "expected": expected_length,
                "observed": len(files_list),
            }

            if results["length"]["observed"] != results["length"]["expected"]:
                message = """Length mismatch in output {output_name} in step {step_name} in pipeline {run_id}. Expected len {a} found {b}.""".format(
                    output_name=output_name,
                    step_name=step_name,
                    run_id=pipeline_run._run_id,
                    b=results["length"]["observed"],
                    a=results["length"]["expected"],
                )
                # logging.error(message)
                results["errors"].append(message)

    return results



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Decorators and utilities for prefixing exception stack traces while obscuring
the exception message itself.
"""


import argparse
from collections.abc import Iterable
import functools
import re
import sys
import time
from traceback import TracebackException

# https://stackoverflow.com/a/38569536
from typing import Callable, Optional, Set, TextIO, Union


PREFIX = "SystemLog:"
SCRUB_MESSAGE = "**Exception message scrubbed**"
LATEST_EXCEPTION = None


class PublicValueError(ValueError):
    """
    Value error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicRuntimeError(RuntimeError):
    """
    Runtime error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicArgumentError(argparse.ArgumentError):
    """
    Argument error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicArgumentTypeError(argparse.ArgumentTypeError):
    """
    Argument type error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicKeyError(KeyError):
    """
    Key error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicTypeError(TypeError):
    """
    Type error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicIndexError(IndexError):
    """
    Index error with public message. Exceptions of this type raised under
    `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will have
    the message prefixed with `PREFIX` in both the printed stack trace and the
    re-raised exception.
    """


class PublicNotImplementedError(NotImplementedError):
    """
    Not implemented error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicFileNotFoundError(FileNotFoundError):
    """
    File not found error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicIOError(IOError):
    """
        I/O error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
        have the message prefixed with `PREFIX` in both the printed stack trace and
        the re-raised exception.
    """


class PublicEnvironmentError(EnvironmentError):
    """
    Environment error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicImportError(ImportError):
    """
    Import error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicStopIteration(StopIteration):
    """
    Stop iteration error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicSystemError(SystemError):
    """
    System error with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


class PublicDeprecationWarning(DeprecationWarning):
    """
    Deprecation warning with public message. Exceptions of this type raised
    under `prefix_stack_trace` or `print_prefixed_stack_trace_and_raise` will
    have the message prefixed with `PREFIX` in both the printed stack trace and
    the re-raised exception.
    """


default_allow_list = [
    PublicValueError.__name__,
    PublicRuntimeError.__name__,
    PublicArgumentError.__name__,
    PublicArgumentTypeError.__name__,
    PublicKeyError.__name__,
    PublicTypeError.__name__,
    PublicIndexError.__name__,
    PublicNotImplementedError.__name__,
    PublicFileNotFoundError.__name__,
    PublicIOError.__name__,
    PublicEnvironmentError.__name__,
    PublicImportError.__name__,
    PublicStopIteration.__name__,
    PublicSystemError.__name__,
    PublicDeprecationWarning.__name__,
]


def _attribute_transformer(prefix: str, scrub_message: str, keep: bool) -> Callable:
    """
    Create a function which may be used to transform exception attributes.

    If an attribute is string-valued or int-valued,
    apply the logic keep? prefix + attr: prefix
    + scrub_message.

    If the attribute is iterable, apply this logic to each member of the
    attribute.

    If the attribute is callable, don't change it.

    If it is neither, replace it with None to ensure no private data leaks.
    """

    def inner(o):
        rv = o
        if isinstance(o, str):
            if keep:
                rv = prefix + o
            else:
                rv = prefix + scrub_message
        elif isinstance(o, int):
            # handling SystemExit
            rv = inner(str(o))
        elif isinstance(o, Iterable):
            rv = type(o)(map(inner, o))  # type: ignore
        elif callable(o):
            rv = rv
        elif not keep:
            rv = None

        return rv

    return inner


def scrub_exception(
    exception: Optional[BaseException],
    scrub_message: str,
    prefix: str,
    keep_message: bool,
    allow_list: list,
    _seen: Optional[Set[int]] = None,
) -> Optional[BaseException]:
    """
    Recursively scrub all potentially private data from an exception, using the
    logic in `_attribute_transformer`.

    Inspired by Dan Schwartz's closed-source implementation:
    https://dev.azure.com/eemo/TEE/_git/TEEGit?path=%2FOffline%2FFocusedInbox%2FComTriage%2Fcomtriage%2Futils%2Fscrubber.py&version=GBcompliant%2FComTriage&_a=content

    which is closely based on the CPython implementation of the
    TracebackException class:
    https://github.com/python/cpython/blob/master/Lib/traceback.py#L478
    """
    if not exception:
        return None

    # Handle loops in __cause__ or __context__ .
    if _seen is None:
        _seen = set()
    _seen.add(id(exception))

    # Gracefully handle being called with no type or value.
    if exception.__cause__ is not None and id(exception.__cause__) not in _seen:
        exception.__cause__ = scrub_exception(
            exception.__cause__,
            scrub_message,
            prefix,
            keep_message,
            allow_list,
            _seen,
        )
    if exception.__context__ is not None and id(exception.__context__) not in _seen:
        exception.__context__ = scrub_exception(
            exception.__context__,
            scrub_message,
            prefix,
            keep_message,
            allow_list,
            _seen,
        )

    keep = keep_message or is_exception_allowed(exception, allow_list)
    transformer = _attribute_transformer(prefix, scrub_message, keep)

    if not keep:
        for attr in dir(exception):
            if attr and not attr.startswith("__"):
                try:
                    value = getattr(exception, attr)
                except AttributeError:
                    # In some cases, e.g. FileNotFoundError, there are attributes
                    # which show up in dir(e), but for which an AttributeError is
                    # thrown when attempting to access the value. See, e.g.:
                    # https://stackoverflow.com/q/47775772 .
                    continue
                try:
                    # If unable to transform or set the attribute, replace the
                    # entire exception since the attribute value is readable, but
                    # we are unable to scrub it.
                    new_value = transformer(value)
                    setattr(exception, attr, new_value)
                except BaseException as e:
                    new_exception = PublicRuntimeError(
                        f"{prefix} Obtained {type(e).__name__} when trying to scrub {attr} from {type(exception).__name__}"  # noqa: E501
                    )
                    new_exception.__cause__ = exception.__cause__
                    new_exception.__context__ = exception.__context__
                    exception = new_exception
                    break

    return exception


def is_exception_allowed(
    exception: Union[BaseException, TracebackException], allow_list: list
) -> bool:
    """
    Check if message is allowed, either by `allow_list`, or `default_allow_list`.

    Args:
        exception (TracebackException): the exception to test
        allow_list (list): list of regex expressions. If any expression matches
            the exception name or message, it will be considered allowed.

    Returns:
        bool: True if message is allowed, False otherwise.
    """
    if not isinstance(exception, TracebackException):
        exception = TracebackException.from_exception(exception)

    # empty list means all messages are allowed
    for expr in allow_list + default_allow_list:
        if re.search(expr, getattr(exception, "_str", ""), re.IGNORECASE):
            return True
        if re.search(expr, getattr(exception.exc_type, "__name__", ""), re.IGNORECASE):
            return True
    return False


def print_prefixed_stack_trace_and_raise(
    file: TextIO = sys.stderr,
    prefix: str = PREFIX,
    scrub_message: str = SCRUB_MESSAGE,
    keep_message: bool = False,
    allow_list: list = [],
    add_timestamp: bool = False,
    err: Optional[BaseException] = None,
) -> None:
    """
    Print the current exception and stack trace to `file` (usually client
    standard error), prefixing the stack trace with `prefix`.
    Args:
        keep_message (bool): if True, don't scrub message. If false, scrub (unless
            allowed).
        allow_list (list): exception allow_list. Ignored if keep_message is True. If
            empty all messages will be srubbed.
        err: the error that was thrown. None accepted for backwards compatibility.
    """
    global LATEST_EXCEPTION
    if err is None:
        err = sys.exc_info()[1]
    scrubbed_err = err
    if err != LATEST_EXCEPTION:
        scrubbed_err = scrub_exception(
            err, scrub_message, prefix, keep_message, allow_list
        )
        LATEST_EXCEPTION = err

    tb_exception = TracebackException.from_exception(scrubbed_err)  # type: ignore

    for execution in tb_exception.format():
        if "return function(*func_args, **func_kwargs)" in execution:
            # Do not show the stack trace for our decorator.
            continue
        for line in execution.splitlines():
            if add_timestamp:
                current_time = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
                print(f"{prefix} {current_time} {line}", file=file)
            else:
                print(f"{prefix} {line}", file=file)

    raise scrubbed_err  # type: ignore


class _PrefixStackTraceWrapper:
    """
    Callable object for catching exceptions and printing their stack traces,
    appropriately prefixed.

    This is an object instead of a nested function to support working in Spark.
    Python anonymous functions are not "pickleable", so we need to define a
    class which handles this logic, so the class is picklable.
    """

    def __init__(
        self,
        file: TextIO,
        disable: bool,
        prefix: str,
        scrub_message: str,
        keep_message: bool,
        allow_list: list,
        add_timestamp: bool,
    ) -> None:
        self.allow_list = allow_list
        self.disable = disable
        self.file = file
        self.keep_message = keep_message
        self.prefix = prefix
        self.scrub_message = scrub_message
        self.add_timestamp = add_timestamp

    def __call__(self, function) -> Callable:
        @functools.wraps(function)
        def wrapper(*func_args, **func_kwargs):
            """
            Create a wrapper which catches exceptions thrown by `function`,
            scrub exception messages, and logs the prefixed stack trace.
            """
            caught_err = None
            try:
                return function(*func_args, **func_kwargs)
            except BaseException as err:
                caught_err = err

            if caught_err:
                print_prefixed_stack_trace_and_raise(
                    self.file,
                    self.prefix,
                    self.scrub_message,
                    self.keep_message,
                    self.allow_list,
                    self.add_timestamp,
                    caught_err,
                )

        return function if self.disable else wrapper


def prefix_stack_trace(
    file: TextIO = sys.stderr,
    disable: bool = bool(sys.flags.debug),
    prefix: str = PREFIX,
    scrub_message: str = SCRUB_MESSAGE,
    keep_message: bool = False,
    allow_list: list = [],
    add_timestamp: bool = False,
) -> Callable:
    """
    Decorator which wraps the decorated function and prints the stack trace of
    exceptions which occur, prefixed with `prefix` and with exception messages
    scrubbed (replaced with `scrub_message`). To use this, just add
    `@prefix_stack_trace()` above your function definition, e.g.

        @prefix_stack_trace()
        def foo(x):
            pass
    """
    return _PrefixStackTraceWrapper(
        file, disable, prefix, scrub_message, keep_message, allow_list, add_timestamp
    )


class PrefixStackTrace:
    def __init__(
        self,
        file: TextIO = sys.stderr,
        disable: bool = bool(sys.flags.debug),
        prefix: str = PREFIX,
        scrub_message: str = SCRUB_MESSAGE,
        keep_message: bool = False,
        add_timestamp: bool = False,
        allow_list: list = [],
    ):
        self.file = file
        self.disable = disable
        self.prefix = prefix
        self.scrub_message = scrub_message
        self.keep_message = keep_message
        self.add_timestamp = add_timestamp
        self.allow_list = allow_list

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type and not self.disable:
            print_prefixed_stack_trace_and_raise(
                file=self.file,
                prefix=self.prefix,
                scrub_message=self.scrub_message,
                keep_message=self.keep_message,
                allow_list=self.allow_list,
                add_timestamp=self.add_timestamp,
                err=exc_value,
            )



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Python utilities that help users to manage, validate
and submit AML pipelines
"""

from .module_helper import AMLModuleLoader
from .pipeline_helper import AMLPipelineHelper

# from .federated_learning import FederatedPipelineBase, StepOutput
from .ray_actor import ray_actor_on_shrike, b64_decode, b64_encode
from shrike.compliant_logging import get_args_from_component_spec

__all__ = [
    "AMLPipelineHelper",
    "AMLModuleLoader",
    # "FederatedPipelineBase",
    # "StepOutput",
    "ray_actor_on_shrike",
    "b64_decode",
    "b64_encode",
    "get_args_from_component_spec",
]



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

"""
Utility methods around unit testing.
"""

import io
import logging
from typing import Union


class StreamHandlerContext:
    """
    Add, then remove a stream handler with the provided format string. The
    `__str__` method on this class returns the value of the internal stream.
    """

    def __init__(self, log: logging.Logger, fmt: str, level):
        self.logger = log

        self._original_level = log.level
        # It's important to set the logger level before adding the handler.
        self.logger.setLevel(level)

        self.stream = io.StringIO()
        self.handler = logging.StreamHandler(self.stream)
        self.handler.setFormatter(logging.Formatter(fmt))
        self.handler.setLevel(level)

    def __enter__(self):
        self.logger.addHandler(self.handler)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.logger.removeHandler(self.handler)
        self.handler.flush()
        self.logger.setLevel(self._original_level)

    def __str__(self):
        return self.stream.getvalue()


def stream_handler(
    log: Union[str, logging.Logger],
    format: str = "%(levelname)s:%(name)s:%(message)s",
    level: Union[None, int, str] = None,
) -> StreamHandlerContext:
    """
    Attach an in-memory stream handler to the provided logger, either a Python
    object or the name of the logger. Can optionally provide log format and
    level.
    """
    if isinstance(log, str):
        log = logging.getLogger(log)

    if not level:
        level = log.getEffectiveLevel()

    return StreamHandlerContext(log, format, level)



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

from abc import ABC, abstractmethod
import base64
import logging
from omegaconf import OmegaConf
from pathlib import Path
import subprocess
import sys
import os
import json
from typing import Any, Dict, List, Optional, Tuple, Union
from dataclasses import asdict

from shrike import __version__
from shrike.build.core.configuration import Configuration, load_configuration
from shrike.build.utils.utils import TelemetryLogger

log = logging.getLogger(__name__)

os.environ["AZURE_ML_CLI_PRIVATE_FEATURES_ENABLED"] = "True"
os.environ["AZURE_ML_INTERNAL_COMPONENTS_ENABLED"] = "True"

class _LogEmphasize:
    def __init__(self, line: str):
        self.line = line

    def __enter__(self):
        log.info(self.line)

    def __exit__(self, exc_type, exc_value, traceback):
        log.info(self.line)


class Command(ABC):
    """
    Commands exposed by this package should subclass this class and implement
    the `run_with_config` method. They should be invoked by calling
    `Subclass().run()` inside their module's `__main__` logic.
    """

    @abstractmethod
    def __init__(self) -> None:
        self.config: Configuration = None  # type: ignore
        self._component_statuses: Dict[str, Dict[str, str]] = {}
        self._errors: List[str] = []
        self.nb_cores = 1

    def attach_workspace(self, workspace_id: str = None) -> None:
        """
        Run `az ml folder attach` to the configured workspace ID. Default to the
        first configured workspace if none is provided.
        """
        working_direcotry = self.config.working_directory
        if workspace_id is None:
            if not self.config.workspaces and self.config.registries:
                workspace_id = self.config.validation_workspace
                if not workspace_id:
                    self.register_error("No workspaces are configured. If you want to publish to registries only, please specify one workspace string in `validation_workspace` for validating components.")
                    return
            else:
                try:
                    workspace_id = self.config.workspaces[0]
                except IndexError:
                    self.register_error(
                        f"No workspaces are configured. Please include them in your configuration file and ensure the path to your configuration file is correct relative to the working directory {working_direcotry} using `--configuration-file PATH/TO/CONFIGURATION_FILE`."
                    )
                    return

        (subscription_id, resource_group, workspace) = self.parse_workspace_arm_id(
            workspace_id
        )
        success = self.execute_azure_cli_command(
            f"account set --subscription {subscription_id}"
        )
        
        # expecting experimental warning in stderr for the following two commands
        # so set stderr_is_failure=False
        success = success and self.execute_azure_cli_command(
            f"config set defaults.group={resource_group} defaults.workspace={workspace}", stderr_is_failure=False, ignore_blowfish = True
        )
        # the following code is needed because
        # the command above won't return any error if attaching an invalid workspace.
        success = success and self.execute_azure_cli_command("ml workspace show", stderr_is_failure=False)
        if not success:
            self.register_error(f"Error!! Failed to attach to {workspace_id}!")

    def display_all_statuses(self) -> None:
        """
        Display all component statuses in an easily readable format.
        """
        pass

    def emphasize(self, line: str = "#" * 80) -> _LogEmphasize:
        """
        Use this to initialize a `with` block for emphasizing any logs inside
        that block.
        """
        return _LogEmphasize(line)

    def ensure_component_cli_installed(self) -> bool:
        """
        Check if the azure CLI is installed;
        install it if not.
        """
        # Removing the v1 version of the azure CLI regardless.
        self.execute_azure_cli_command(command="extension remove -n azure-cli-ml", stderr_is_failure=False)

        # Check whether the azure CLI is installed
        component_cli_exists = self.execute_azure_cli_command(
            "ml component -h",
            stderr_is_failure=False,
            log_error=False,
        )

        if component_cli_exists:
            log.info("azure CLI exists. Skipping installation.")
            return True
        else:
            log.info(
                f"installing latest azure CLI."
            )
            self.execute_azure_cli_command(command="extension remove -n ml", stderr_is_failure=False,)

            cli_install_command = f"extension add -n ml --yes"
            if self.config.verbose:
                cli_install_command += " --verbose"

            is_installed = self.execute_azure_cli_command(
                command=cli_install_command,
                # installation may show time to install
                stderr_is_failure=False,
            )

            if is_installed:
                log.info("azure CLI is installed.")
            else:
                log.error("azure CLI installation failed.")

            return is_installed

    def execute_azure_cli_command(
        self,
        command: str,
        working_dir: Optional[str] = None,
        stderr_is_failure: bool = True,
        fail_if_version_exists: bool = False,
        log_error: bool = True,
        check_stdout_result: bool = False,
        ignore_blowfish: bool = False,
    ) -> bool:
        """
        Use this method, NOT `execute_command`, for running Azure CLI commands.
        The `command` string should contain everything AFTER the `az`.

        This does NOT use the `azure-cli-core` Python package
        ( https://stackoverflow.com/a/55960725 ) because it takes a long time
        to install, and does not work in Windows.

        This method is necessary for subtle reasons around the way Azure CLI
        exposes commands. The "naive approach" doesn't work.
        """
        log.debug(f"Executing: az {command}")
        az_command_bytes = bytes(f"az {command}", "utf-16le")
        az_command_b64 = base64.b64encode(az_command_bytes).decode("ascii")
        pwsh_command = ["pwsh", "-EncodedCommand", az_command_b64]
        success = self.execute_command(
            pwsh_command, working_dir, stderr_is_failure, fail_if_version_exists, log_error, check_stdout_result, ignore_blowfish
        )
        return success

    def execute_command(
        self,
        command: List[str],
        working_dir: Optional[str] = None,
        stderr_is_failure: bool = True,
        fail_if_version_exists: bool = False,
        log_error: bool = True,
        check_stdout_result: bool = False,
        ignore_blowfish: bool = False,
    ) -> bool:
        """
        Execute the provided shell command using the configured timeout. Working
        directory defaults to the configured one. If `stderr_is_failure` is
        set to false, stderr from the command will be converted to "vanilla"
        logs and will not affect success; 

        Logs are NOT streamed realtime - they are "bundled together" after the
        command executes or times out.

        Warning: running `az *` naively via this function will not work, since
        the Azure CLI is not, by default, discoverable via `subprocess.run`.
        """
        if working_dir is None:
            working_dir = self.config.working_directory

        if len(command) > 0 and command[0] == "az":
            raise ValueError(
                "Do not run Azure CLI commands with this function. Use execute_azure_cli_command instead."
            )

        kwargs = {}

        if stderr_is_failure or fail_if_version_exists:
            kwargs["stderr"] = subprocess.PIPE

        log.debug(f"Executing {command} in {working_dir}")

        timeout = self.config.shell_command_timeout_in_seconds

        try:
            res = subprocess.run(
                args=command,
                cwd=working_dir,
                stdout=subprocess.PIPE,
                timeout=timeout,
                **kwargs,
            )
            success = res.returncode == 0

            if not success:
                if log_error:
                    log.error(f"Command failed with exit code {res.returncode}")
                else:
                    log.info(f"Command failed with exit code {res.returncode}")

            stdout = res.stdout
            stderr = res.stderr
            if success and check_stdout_result:
                # https://msdata.visualstudio.com/Vienna/_workitems/edit/2048751
                # success = input of operation (`validate`) is valid
                # but need to check the results in stdout
                if isinstance(stdout, bytes):
                    stdout = stdout.decode("utf-8")
                success = json.loads(stdout)["result"] == "Succeeded"
        except subprocess.TimeoutExpired as e:
            log.error(f"Command timed out after {timeout} seconds.")
            success = False
            stdout = e.stdout
            stderr = e.stderr
        
        if stdout:
            for line in stdout.splitlines():
                try:
                    line = str(line, encoding="utf-8", errors="ignore")  # type: ignore
                except:
                    log.debug(
                        "Failed to convert the following stdout line into String (utf-8)"
                    )
                log.info(line)
        if stderr:
            stderr = stderr.decode() if isinstance(stderr, bytes) else stderr
            if ignore_blowfish and "Blowfish" in stderr:
                log.info("Ignoring Blowfish error.")
                return True
                
            for line in stderr.splitlines():
                try:
                    line = str(line, encoding="utf-8", errors="ignore")  # type: ignore
                except:
                    log.debug(
                        "Failed to convert the following stdout line into String (utf-8)"
                    )
                if stderr_is_failure:
                    log.error(line)
                    success = False
                elif fail_if_version_exists and "Error" in line and "already exists in" in line:
                    log.error(line)
                    success = False                       
                else:
                    log.info(line)

        return success

    def normalize_path(self, path: Union[str, Path], directory=False) -> str:
        """
        Normalize the provided path (file or directory) to the following format:
        - Absolute (not relative)
        - Linux-style (forward slash separating directories)
        - If `directory=True`, ending in a forward slash.
        """
        if isinstance(path, str):
            path = Path(path)

        path = str(path.absolute())
        rv = path.replace("\\", "/")

        if directory and not rv[-1] == "/":
            rv += "/"

        return rv

    def parse_workspace_arm_id(self, id: str) -> Tuple[str, str, str]:
        """
        Parse a workspace ARM ID like
        `/subscriptions/df3f8bd1-3b4f-4c33-93e0-187f162066d2/resourceGroups/aml1p-rg/providers/Microsoft.MachineLearningServices/workspaces/aml1p-ml-wus2`
        and return (subscription ID, resource group, workspace name).
        """
        split = id.split("/")
        subscription = split[2]
        resource_group = split[4]
        workspace = split[8]
        return (subscription, resource_group, workspace)

    def register_component_status(
        self, component_name: str, status_name: str, status: str
    ) -> None:
        """
        Register a status (e.g., build = failed) for a specified component. All
        statuses will be displayed in a friendly manner before exiting.
        """
        if component_name not in self._component_statuses:
            self._component_statuses[component_name] = {}

        status_dict = self._component_statuses[component_name]
        status_dict[status_name] = status

    def register_error(self, error: str) -> None:
        """
        Register that an error has occured (also, log it). If any errors have
        been registered, the `run` method will return with non-zero exit code.
        """
        log.error(error)
        self._errors.append(error)

    def telemetry_logging(self, command: str) -> None:
        """
        Log the telemetry information in the Azure Application Insights
        """
        telemetry_logger = TelemetryLogger(
            enable_telemetry=not self.config.disable_telemetry
        )
        telemetry_logger.log_trace(
            message=f"shrike.build=={__version__}: {command}",
            properties={
                "custom_dimensions": {"configuration": str(asdict(self.config))}
            },
        )

    def run(self) -> None:
        """
        Call this to load the configuration object, initialize the logging tree,
        then invoke the subclass' `run_with_config` method and return the
        appropriate exit code. This should be the entrypoint inside a command's
        `if __name__ == "__main__"` block.
        """
        config = load_configuration()

        log_level = "DEBUG" if config.verbose else "INFO"
        logging.basicConfig(level=log_level, format=config.log_format)
        
        max_nb_cores = max(os.cpu_count() - 1, 1) # type: ignore
        if config.number_of_cores_parallel <= 0 or (config.number_of_cores_parallel > max_nb_cores and not config.force_parallelism):
            self.nb_cores = max_nb_cores
        else:
            self.nb_cores = config.number_of_cores_parallel

        with self.emphasize():
            config_yaml = OmegaConf.to_yaml(config)
            log.info("Final configuration being used:\n")
            log.info(config_yaml)

        self.config = config
        self.run_with_config()

        self.display_all_statuses()

        failed = bool(self._errors)

        if failed:
            log.error(f"Encountered {len(self._errors)} errors!")

        sys.exit(bool(self._errors))

    @abstractmethod
    def run_with_config(self) -> None:
        """
        Run the subclasses command with the specified configuration object.
        Before this method is invoked, there is no guarantee that `self.config`
        will be populated; after it is invoked, that is guaranteed.
        Implementations of this method should NOT mutate the logging tree in
        any way. They should also NOT raise any exceptions; rather they should
        call the `register_error` method, which will ensure non-zero exit code.
        Implementations can raise specific "status information" (e.g., a
        component is not "active") by calling `register_component_status`.
        """
        pass



# Copyright (c) Microsoft Corporation.
# Licensed under the MIT license.

import logging
from opencensus.ext.azure.log_exporter import AzureLogHandler

log = logging.getLogger(__name__)


class TelemetryLogger:
    """Utils class for opencensus azure monitor"""

    def __init__(
        self, enable_telemetry=True, instrumentation_key=None, level=logging.INFO
    ):
        self.logger = logging.getLogger("telemetry_logger")
        self.logger.setLevel(level)
        self.enable_telemetry = enable_telemetry
        # Why is it okay to include this key directly in the source code?
        # For any client-side tool, there is a fundamental problem with protecting instrumentation
        # keys. You want the published tool to be able to collect telemetry, but the only way
        # it can do this is if it has some kind of instrumentation key.
        #
        # For an authoritative example, the dotnet CLI contains their telemetry key in a
        # public GitHub repository:
        # https://github.com/dotnet/cli/blob/master/src/dotnet/Telemetry/Telemetry.cs
        #
        # The underlying Azure resource is called `aml1p-ml-tooling`.
        self.instrumentation_key = (
            "3b7f42bf-028a-4251-9965-b4387e9d6617"
            if instrumentation_key is None
            else instrumentation_key
        )
        handler = AzureLogHandler(
            connection_string=f"InstrumentationKey={self.instrumentation_key}"
        )
        handler.add_telemetry_processor(self.scrubber_function)
        self.logger.addHandler(handler)

    def log_trace(self, message, properties={}, level=logging.INFO):
        if self.enable_telemetry:
            try:
                if level == logging.INFO:
                    self.logger.info(message, extra=properties)
                elif level == logging.WARNING:
                    self.logger.warning(message, extra=properties)
                elif level == logging.ERROR:
                    self.logger.error(message, extra=properties)
                elif level == logging.CRITICAL:
                    self.logger.critical(message, extra=properties)
                else:
                    log.error("The logging level is not expected!")
            except Exception as ex:
                log.warning("Send telemetry exception: %s", str(ex))
        else:
            log.info(
                "Sending trace log messages to application insight has been disabled."
            )

    # Callback function to scrub some columns
    def scrubber_function(self, envelope):
        envelope.tags["ai.cloud.roleInstance"] = "cloud_RoleInstance_Scrubbed"
        envelope.tags["ai.location.ip"] = "IP_Scrubbed"
